import os
import hashlib
from pathlib import Path

import MDAnalysis as mda
import numpy as np
import pytest

import chilife as xl

U = xl.load_protein("test_data/m1omp.pdb")
protein = U.select_atoms("protein")

load_protein_args = [('test_data/1ubq.pdb',),
                     (Path('test_data/1ubq.pdb'),),
                     ('test_data/traj_io.pdb', Path('test_data/traj_io.xtc')),]


@pytest.mark.parametrize('args', load_protein_args)
def test_load_protein(args):

    struct = xl.load_protein(*args)
    assert isinstance(struct, mda.Universe)

    
    if len(args) == 2:
        if (p1 := Path('test_data/.traj_io.xtc_offsets.lock')).exists():
            os.remove('test_data/.traj_io.xtc_offsets.lock')
        if (p2 := Path('test_data/.traj_io.xtc_offsets.npz')).exists():
            os.remove('test_data/.traj_io.xtc_offsets.npz')


def test_save():
    L20R1 = xl.SpinLabel("R1C", 20, protein)
    S238T = xl.RotamerEnsemble("THR", 238, protein, eval_clash=False)
    A318DHC = xl.dSpinLabel("DHC", [318, 322], protein, rotlib='test_data/DHC')

    xl.save(L20R1, S238T, A318DHC, protein, KDE=False)

    with open(f"test_data/test_save.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("No_Name_Protein_many_labels.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove("No_Name_Protein_many_labels.pdb")

    assert ans == test


def test_save_frame():
    U = mda.Universe('test_data/traj_io.pdb', 'test_data/traj_io.xtc')
    xl.save('tmp.pdb', U, frames=5)

    with open(f"test_data/save_frame.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("tmp.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('tmp.pdb')

    assert ans == test


def test_save_ic_frame():
    U = mda.Universe('test_data/xlsavetraj.pdb')
    sele_IC = xl.MolSysIC.from_atoms(U.atoms)
    xl.save('ic_frames.pdb', sele_IC)

    with open(f"test_data/ic_frames.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    with open("ic_frames.pdb", "r") as f:
        test = hashlib.md5(f.read().encode("utf-8")).hexdigest()

    os.remove('ic_frames.pdb')
    assert ans == test


def test_save_fail():
    with pytest.raises(TypeError):
        xl.save("tmp", np.array([1, 2, 3]))


def test_save_multiple_groups():
    CAs = U.select_atoms('name CA')
    CBs = U.select_atoms('name CB')

    xl.save(U.atoms, CAs, CBs)

    names = ['m1omp', 'm1omp1', 'm1omp2']

    with open(f"No_Name_Protein.pdb", "r") as f:
        for line in f:
            for name in iter(names):
                sen = f"HEADER {name}\n"
                if line == sen:
                    names.remove(name)
                if len(names) ==0:
                    break
            if len(names) == 0:
                break

    assert len(names) == 0

    os.remove('No_Name_Protein.pdb')


def test_write_bonds():

    bonds = [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [1, 4], [1, 5], [1, 6], [2, 3], [2, 4], [2, 5], [2, 6]]
    with open('test_write_bonds.pdb', 'w') as f:
        xl.write_bonds(f, bonds)

    with open('test_write_bonds.pdb', 'r') as f:
        lines = "".join(f.readlines())
        thash = hashlib.md5(lines.encode("utf-8")).hexdigest()

    with open('test_data/test_write_bonds.pdb', 'r') as f:
        lines = "".join(f.readlines())
        ahash = hashlib.md5(lines.encode("utf-8")).hexdigest()

    os.remove('test_write_bonds.pdb')
    assert ahash == thash


def test_write_protein_with_bonds():
    prot1 = xl.load_protein('test_data/alphabetical_peptide.pdb')
    R1M = xl.SpinLabel('R1M')
    prot1.add_bonds(xl.guess_bonds(prot1.atoms.positions, prot1.atoms.types))

    xl.save(R1M, conect=True)


    with open('1R1M.pdb', 'r') as f:
        lines = "".join([line for line in f if line.startswith("CONECT")])
        thash = hashlib.md5(lines.encode("utf-8")).hexdigest()

    with open('test_data/cnct.pdb', 'r') as f:
        lines = "".join([line for line in f if line.startswith("CONECT")])
        ahash = hashlib.md5(lines.encode("utf-8")).hexdigest()

    assert thash == ahash

    os.remove('1R1M.pdb')


from pathlib import Path
import rtoml
import pickle
from collections import defaultdict
import numpy as np

# Define useful global variables
SUPPORTED_BB_LABELS = ("R1C",)
DATA_DIR = Path(__file__).parent.absolute() / "data/"
RL_DIR = Path(__file__).parent.absolute() / "data/rotamer_libraries/"

with open(RL_DIR / 'additional_rotlib_dirs.txt', 'r') as f:
    USER_RL_DIR = [Path(x.strip()) for x in f.readlines()]

with open(RL_DIR / "defaults.toml", "r") as f:
    rotlib_defaults = rtoml.load(f)

USER_LIBRARIES = {f.name[:-11] for f in (RL_DIR / "user_rotlibs").glob("*rotlib.npz")}
USER_dLIBRARIES = {f.name[:-12] for f in (RL_DIR / "user_rotlibs").glob("*drotlib.zip")}
USER_dLIBRARIES = USER_dLIBRARIES | {f.name[:-15] for f in (RL_DIR / "user_rotlibs").glob("*drotlib.zip")}

DATA_DIR = Path(__file__).parent.absolute() / "data/"
RL_DIR = Path(__file__).parent.absolute() / "data/rotamer_libraries/"

# Define rotamer dihedral angle atoms
with open(DATA_DIR / "dihedral_defs.toml", "r") as f:
    dihedral_defs = rtoml.load(f)

SUPPORTED_RESIDUES = set(dihedral_defs.keys())
[SUPPORTED_RESIDUES.remove(lab) for lab in ("CYR1", "MTN")]

with open(RL_DIR / "RotlibIndexes.pkl", "rb") as f:
    rotlib_indexes = pickle.load(f)

with open(DATA_DIR / 'BondDefs.pkl', 'rb') as f:
    bond_hmax_dict = {key: (val + 0.4 if 'H' in key else val + 0.37) for key, val in pickle.load(f).items()}
    bond_hmax_dict = defaultdict(lambda: 0, bond_hmax_dict)


    def bond_hmax(a): return bond_hmax_dict.get(tuple(i for i in a), 0)


    bond_hmax = np.vectorize(bond_hmax, signature="(n)->()")

atom_order = {"N": 0, "CA": 1, "C": 2, "O": 3, 'H': 5}

nataa_codes = {'ALA': 'A', 'CYS': 'C', 'ASP': 'D', 'GLU': 'E', 'PHE': 'F', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I',
               'LYS': 'K', 'LEU': 'L', 'MET': 'M', 'ASN': 'N', 'PRO': 'P', 'GLN': 'Q', 'ARG': 'R',
               'SER': 'S', 'THR': 'T', 'VAL': 'V', 'TRP': 'W', 'TYR': 'Y'}



natnu_codes = {'A', 'C', 'G', 'I', 'U', 'DA', 'DC', 'DG', 'DI', 'DT', 'DU'}

mm_backbones = {'aa': ['N', 'CA', 'C', 'O'],
                'nu': ["P",  "O5'", "C5'",  "C4'", "O4'", "C3'",  "O3'", "C2'", "O2'", "C1'",  "OP1", "OP2"],
                'ACE': ['C', 'O', 'CH3']}


alt_prot_states = {'ASH': 'ASP', 'GLH': 'GLU', 'LYN': 'LYS', 'HID': 'HIS', 'HIE': 'HIS', 'HIP': 'HIS', 'CYM': 'CYS'}
ralt_prot_states = {
    'ASP': {13: 'ASH', 12: 'ASP'},
    'GLU': {16: 'GLH', 15: 'GLU'},
    'LYS': {21: 'LYN', 22: 'LYS'},
    'HIS': {17: 'HIS', 18: 'HIP'},
    'CYS': {10: 'CYM', 11: 'CYS'}
    }


inataa = {val: key for key, val in nataa_codes.items() if key not in ('CME', 'MSE')}
nataa_codes.update(inataa)
natnu_codes.update({'MSE': 'M', 'CME': 'C', 'HID': 'H', 'HIP': 'H', 'HIE': 'H',
                    'LYN': 'K', 'CYM': 'C', 'ASH': 'D', 'GLH': 'E'})
del inataa


# Define rotamer dihedral angle atoms
with open(RL_DIR / "ncaps.txt", "r") as f:
     ncaps = tuple(f.readline().split())

with open(RL_DIR / "ccaps.txt", "r") as f:
    ccaps = tuple(f.readline().split())






import itertools
import logging
import warnings
from collections import defaultdict
from typing import List, Union, Dict, Tuple

import MDAnalysis
import igraph as ig
import numpy as np
from numpy.typing import ArrayLike

from .MolSys import MolecularSystemBase, Trajectory, MolSys
from .Topology import Topology, guess_bonds
from .protein_utils import get_angles, get_dihedrals
from .numba_utils import _ic_to_cart, batch_ic2cart
from .globals import dihedral_defs


class MolSysIC:
    """
    A class for protein internal coordinates.
    
    Parameters
    ----------
    z_matrix : np.ndarray
        Array of z-matricies for all frames of the ensemble/trajectory
    z_matrix_idxs : np.ndarray
        Indices of the attoms that define the bond lengths, angles, and dihedrals of the z-matrix
    protein :
    kwargs

    Attributes
    ----------
    zmats : dict[np.ndarray]
        Dictionary of Z-matrices of the molecule. Each entry corresponds to a contiguous segment of bonded atoms.
    z_matrix_idxs : dict[np.ndarray]
        Dictionary of Z-matrix indices.
    atom_dict : dict
        Nested dictionaries containing indices of the atoms that define the coordinates (bond, angle dihedral) objects.
        Nesting pattern: [coord_type][chain][resi, stem][name] where coord_type is one of 'bond', 'angle' or 'dihedral',
        chain is the chainid, resi is the residue number, stem is a tuple of the names of the 3 atoms preceding the atom
        coordinate being defined.
    ICS : dict
        Nested dictionaries containing AtomIC objects. Nesting pattern: [chain][resi][stem][name] where chain is the
        chainid, resi is the residue number, stem is a tuple of the names of the 3 atoms preceding the atom coordinate
        being defined.
    atoms : np.ndarray[AtomIC]
        List of AtomIC objects making up the protein in order of their indices.
    atom_types : np.ndarray
        Array of atom types (elements) corresponding to the atom indices.
    atom_names : np.ndarray
        Array of atom names corresponding to the atom indices.
    resis : np.ndarray
        Ordered array of residue numbers.
    resnames : dict[int]:str
        Dictionary mapping residue numbers to their residue names.
    chains : np.ndarray
        Array of chainids
    bonded_pairs : np.ndarray
        Array of index pairs corresponding to atoms that are bonded.
    perturbed : bool
        Indicates if any internal coordinates have changed since the last time cartesian coordinates were calculated.
    dihedral_defs : list
        List of major side chain and backbone dihedral definitions as defined by user defined and supported residues.
    """

    def __init__(self,
                 z_matrix: ArrayLike,
                 z_matrix_idxs: ArrayLike,
                 protein: Union[MolecularSystemBase, MDAnalysis.Universe, MDAnalysis.AtomGroup],
                 **kwargs):
        # Internal coords and atom info
        if isinstance(protein, MolSys):
            self.protein = protein.copy()
        else:
            self.protein = MolSys.from_atomsel(protein)

        self.atoms = self.protein.atoms
        self.atom_names = self.protein.names
        self.atom_chains = self.protein.segids
        self.atom_resnums = self.protein.resnums
        self.atom_types = self.protein.types
        self.atom_resnames = self.protein.resnames
        uresid, uresidx = np.unique(self.protein.resixs, return_index=True)
        self.resnums = self.atom_resnums[uresidx]
        self.resnames = self.atom_resnames[uresidx]
        self.trajectory = Trajectory(z_matrix, self)
        self.z_matrix_idxs = z_matrix_idxs
        self.chain_operator_idxs = kwargs.get('chain_operator_idxs', None)
        self.has_chain_operators = bool(kwargs.get('chain_operators', None))

        self.chains = np.unique(self.atom_chains)
        self._chain_segs = [[a, b] for a, b in zip(self.chain_operator_idxs,
                                                   self.chain_operator_idxs[1:] + [len(self.z_matrix_idxs)])]

        if not self.has_chain_operators:
            self.chain_operators = None
        else:
            self._chain_operators = kwargs['chain_operators']


        # Topology
        self.bonds = kwargs['bonds'] if 'bonds' in kwargs else \
            guess_bonds(self.protein.positions, self.protein.types)
        self._nonbonded = kwargs.get('nonbonded', None)
        self.topology = kwargs['topology'] if 'topology' in kwargs else \
            Topology(self.protein, self.bonds)

        self.non_nan_idxs = kwargs.get('non_nan_idxs', None)
        if self.non_nan_idxs is None:
            self.non_nan_idxs = np.argwhere(~np.any(self.z_matrix_idxs < 0, axis=1)).flatten()

        self.nan_idxs = np.argwhere(np.any(self.z_matrix_idxs < 0, axis=1)).flatten()

        self.chain_res_name_map = kwargs.get('chain_res_name_map', defaultdict(list))
        if self.chain_res_name_map == {}:
            idxs, b2s, b1s,  _ = self.z_matrix_idxs[self.non_nan_idxs].T
            chains = self.atoms[b2s].segids
            resnums = self.atoms[b2s].resnums
            [self.chain_res_name_map[(chain, res, b1, b2)].append(idx)
             for chain, res, b1, b2, idx in
             zip(chains, resnums, self.atom_names[b1s], self.atom_names[b2s], idxs)]

            self.chain_res_name_map = {k: v for k, v in self.chain_res_name_map.items()}

        # Misc
        self.perturbed = False
        self._dihedral_defs = None

    @classmethod
    def from_atoms(cls,
                   atoms: Union[MDAnalysis.Universe, MDAnalysis.AtomGroup, MolecularSystemBase],
                   preferred_dihedrals: List = None,
                   bonds: ArrayLike = None,
                   **kwargs: Dict):
        """
        Generate a MolSysIC object from a :class:`~chilife.AtomSelection` or `MDAnalysis.AtomGroup` object.

        Parameters
        ----------
        atoms : MDAnalysis.Universe, MDAnalysis.AtomGroup, MolecularSystemBase
            MDA Universe, AtomGroup or chiLife Molecular System to create the internal coordinates from.
        preferred_dihedrals : List[List[str]]
            List of preferred dihedral definitions to use when defining the internal coordinate system.
        bonds : ArrayLike
            Array of tuples defining the atom pairs that are bonded.
        kwargs : dict
            Additional keyword arguments.

            ignore_water : bool
                Ignore atoms that belong to water molecules.
            use_chain_operators : bool
                Allow for the use of a translation and rotation vectors to orient cains that are not covalently linked.

        Returns
        -------
        cls : MDAnalysis.Universe
            A chiLife internal coordinates molecular system.
        """

        if isinstance(atoms, MDAnalysis.Universe):
            atoms = atoms.atoms

        if kwargs.get('ignore_water', True):
            water_atoms = atoms.select_atoms("byres (name OH2 or resname HOH)")
            atoms = atoms[~np.isin(atoms.ix, water_atoms.ix)]

        # Keep track of atom indexes in parent protein object in case atoms come from a larger system
        atom_idxs = atoms.ix

        # Explicitly passing bonds means no other bonds will be used and that the bonds are using universe indices.
        if bonds is not None:
            bonds = bonds.copy()
            ibonds = bonds[np.all(np.isin(bonds, atom_idxs), axis=1)]
            ixmap = {ix: i for i, ix in enumerate(atom_idxs)}

            # renumber bonds to current selection
            bonds = np.vectorize(ixmap.get)(ibonds)
            bonds = np.sort(bonds, axis=1)

        # # Otherwise use bonds defined by the protein object
        # elif hasattr(protein, 'bonds'):
        #     bonds = protein.bonds
        # Guess the bonds as a last resort
        else:
            # Add selection's lowest index in case the user is selecting a subset of atoms from a larger system.
            bonds = guess_bonds(atoms.positions, atoms.types)

        # Remove bonds outside of selection
        topology = Topology(atoms, bonds)
        z_matrix_dihedrals = topology.get_zmatrix_dihedrals()
        # z_mat_map = {k[-1]: i for i, k in enumerate(z_matrix_dihedrals)}

        if preferred_dihedrals is not None:
            present = False
            for dihe in preferred_dihedrals:

                # Get the indices of the atom being defined by the preferred dihedral
                idx_of_interest = np.argwhere(atoms.names == dihe[-1]).flatten()
                for idx in idx_of_interest:

                    # Skip dihedral defs of chain terminal atoms
                    if len(z_matrix_dihedrals[idx]) != len(dihe):
                        continue

                    # Check if it is already in use
                    if np.all(atoms[z_matrix_dihedrals[idx]].names == dihe):
                        present = True
                        continue

                    if idx not in topology.dihedrals_by_atoms:
                        dihedral = [idx]
                        continue

                    # Check for alternative dihedral definitions that satisfy the preferred dihedral
                    for p in topology.dihedrals_by_atoms[idx]:
                        if np.all(atoms[list(p)].names == dihe):
                            dihedral = [a for a in p]
                            break
                    else:
                        dihedral = [idx]

                    # If an alternative is found, replace it in the dihedral list.
                    if len(dihedral) == 4:
                        present = True
                        z_matrix_dihedrals[idx] = dihedral

                        # also change any other dependent dihedrals
                        for dihe_idxs in topology.dihedrals_by_bonds[tuple(dihedral[1:3])]:
                            zmidx = dihe_idxs[-1]
                            tmp = z_matrix_dihedrals[zmidx]
                            if all(a == b for a, b in zip(tmp[1:3], dihedral[1:3])):
                                z_matrix_dihedrals[zmidx][0] = dihedral[0]

            if not present and preferred_dihedrals != []:
                raise ValueError(f'There is no dihedral `{dihe}` in the provided protein. Perhaps there is typo or the '
                                 f'atoms are not sorted correctly')

        z_matrix_coordinates = np.zeros((len(atoms.universe.trajectory), len(z_matrix_dihedrals), 3))
        z_matrix_dihedrals = zmatrix_idxs_to_local(z_matrix_dihedrals)

        chain_operator_idxs = get_chainbreak_idxs(z_matrix_dihedrals)

        chain_operators = []
        for i, ts in enumerate(atoms.universe.trajectory):
            z_matrix, chain_operator = get_z_matrix(atoms.positions, z_matrix_dihedrals, chain_operator_idxs)
            z_matrix_coordinates[i] = z_matrix
            chain_operators.append(chain_operator)

        if not kwargs.get('use_chain_operators', True):
            chain_operators = None

        return cls(z_matrix_coordinates, z_matrix_dihedrals, atoms,
                   chain_operators=chain_operators, chain_operator_idxs=chain_operator_idxs)

    def copy(self):
        """Create a deep copy of an MolSysIC instance"""
        z_matrix = self.trajectory.coordinate_array.copy()
        z_matrix_idxs = self.z_matrix_idxs.copy()
        if isinstance(self._chain_operators, list):
            chain_operators = [{k: {k2: v2.copy() for k2, v2 in v.items()} for k, v in co.items()} for co in self._chain_operators]
        else:
            chain_operators = {k: {k2: v2.copy() for k2, v2 in v.items()} for k, v in self._chain_operators.items()}

        kwargs = {'chain_operators': chain_operators,
                  'chain_operator_idxs': self._chain_operator_idxs,
                  'bonds': self.bonds,
                  'nonbonded': self._nonbonded,
                  'topology': self.topology,
                  'protein': self.protein,
                  'non_nan_idxs': self.non_nan_idxs,
                  'chain_res_name_map': self.chain_res_name_map}

        return MolSysIC(z_matrix, z_matrix_idxs, **kwargs)

    @property
    def chain_operator_idxs(self):
        """An array of indices defining the starting and ending indices of the different chains."""
        return self._chain_operator_idxs

    @chain_operator_idxs.setter
    def chain_operator_idxs(self, val):
        if val is None:
            self._chain_operator_idxs  = get_chainbreak_idxs(self.z_matrix_idxs)
        else:
            self._chain_operator_idxs = val

    @property
    def chain_operators(self):
        """
        dict: A set of coordinate transformations that can orient multiple chains that are not covalently linked. e.g.
        structures with missing residues or protein complexes. The ``chain_operator`` property is a dictionary mapping
        chainIDs to a sub-dictionary containing a translation vector, ``ori``  and rotation matrix ``mx`` that will
        transform the protein coordinates from tie internal coordinate frame to some global frame.
        """
        if isinstance(self._chain_operators, dict):
            return self._chain_operators
        else:
            return self._chain_operators[self.trajectory.frame]

    @chain_operators.setter
    def chain_operators(self, op: Dict):
        """Assign or calculate operators for all chains.

        Parameters
        ----------
        op : dict
            Dictionary containing an entry for each chain in the MolSysIC molecule. Each entry must contain a
            rotation matrix, 'mx' and translation vector 'ori'.
        """

        error_message = 'chain operators must be set to \n' \
                        '    1) a dict to set the chain operators for a single frame \n' \
                        '    2) a list of dicts the same length of the trajectory to set chain operators' \
                        'for each frame\n' \
                        '    3) a list with one dict to set all all frames to the same chain operator.'

        if op is None:
            if hasattr(self, '_chain_operators'):
                from_list = True if isinstance(self._chain_operators, list) else False
            else:
                from_list=False

            logging.info("No protein chain origins have been provided. All chains will start at [0, 0, 0]")

            op = {idx: {"ori": np.array([0, 0, 0]), "mx": np.identity(3)}for idx in self._chain_operator_idxs}
            self.has_chain_operators = False
            self._chain_operators = op
            self.apply_chain_operators(from_list=from_list)
        else:
            self.has_chain_operators = True
            if isinstance(op, dict):
                self._chain_operators[self.trajectory.frame] = op
                self.apply_chain_operators(self.trajectory.frame)
            elif not isinstance(op, list):
                raise RuntimeError(error_message)
            elif len(op) == 1:
                self._chain_operators = op[0]
                self.apply_chain_operators()
            elif len(op) == len(self.trajectory):
                self._chain_operators = op
                self.apply_chain_operators()
            else:
                raise RuntimeError(error_message)

    @property
    def z_matrix(self):
        """The z-matrix (internal) coordinates of the molecular system for the current frame."""
        return self.trajectory.coordinate_array[self.trajectory.frame]

    @property
    def coords(self):
        """The cartesian coordinates of the molecular system for the current frame"""
        self.protein.trajectory[self.trajectory.frame]
        if (self.protein.positions is None) or self.perturbed:
            self.protein.positions = self.to_cartesian()
            self.perturbed = False

        return self.protein.positions

    @coords.setter
    def coords(self, val):
        self.protein.trajectory[self.trajectory.frame]
        self.protein.positions = val
        z_matrix, chain_operator = get_z_matrix(val, self.z_matrix_idxs, self.chain_operator_idxs)

        self.trajectory.coordinate_array[self.trajectory.frame] = z_matrix
        self.chain_operators[self.trajectory.frame] = chain_operator

    @property
    def nonbonded(self):
        """np.ndarray: Array of atom index pairs of atoms that are not bonded"""
        if self._nonbonded is None and not (self.bonds is None or self.bonds.any() is None):
            bonded_pairs = {(a, b) for a, b in self.bonds}
            possible_bonds = itertools.combinations(range(len(self.atoms)), 2)
            self._nonbonded = np.fromiter(
                itertools.chain.from_iterable(
                    nb for nb in possible_bonds if nb not in bonded_pairs), dtype=int)

            self._nonbonded.shape = (-1, 2)

        return self._nonbonded

    def to_cartesian(self):
        """Method to convert z-matrix (internal) coordinate to cartesian coordinate"""
        cart_coords = _ic_to_cart(self.z_matrix_idxs[:, 1:], self.z_matrix)

        # Apply chain operations if any exist
        if self.has_chain_operators:
            for start, end in self._chain_segs:
                op = self.chain_operators[start]
                cart_coords[start:end] = cart_coords[start:end] @ op["mx"] + op["ori"]

        return cart_coords

    def set_dihedral(
            self,
            dihedrals: Union[float, ArrayLike],
            resi: int,
            atom_list: ArrayLike,
            chain: Union[int, str] = None
    ):
        """Set one or more dihedral angles of a single residue in internal coordinates for the atoms defined in atom
        list.

        Parameters
        ----------
        dihedrals : float, ArrayLike
            Angle or array of angles to set the dihedral(s) to.
        resi : int
            Residue number of the site being altered
        atom_list : ArrayLike
            Names or array of names of atoms involved in the dihedral(s)
        chain : int, str, optional
            Chain containing the atoms that are defined by the dihedrals that are being set. Only necessary when there
            is more than one chain in the proteinIC object.

        Returns
        -------
        self : MolSysIC
            ProteinIC object with new dihedral angle(s)
        """

        self.perturbed = True

        chain = self._check_chain(chain)

        dihedrals = np.atleast_1d(dihedrals)
        atom_list = np.atleast_2d(atom_list)
        for i, (dihedral, atoms) in enumerate(zip(dihedrals, atom_list)):
            if (tag := (chain, resi, atoms[1], atoms[2])) not in self.chain_res_name_map:
                raise RuntimeError(f'{atoms} is not a recognized dihedrals of residue {resi}. Please make sure you '
                                   f'have the correct residue number and atom names.')

            pert_idxs = self.chain_res_name_map[tag]  + self.chain_res_name_map.get((chain, resi, atoms[2], atoms[1]), [])
            for idx in pert_idxs:
                protein_atom_names = self.z_matrix_names[idx][::-1]
                if tuple(protein_atom_names) == tuple(atoms):
                    delta = self.z_matrix[idx, 2] - dihedral
                    break
            else:
                raise RuntimeError('')

            self.trajectory.coords[self.trajectory.frame, pert_idxs, 2] -= delta

        return self

    def batch_set_dihedrals(
            self,
            idxs: ArrayLike,
            dihedrals: ArrayLike,
            resi: int,
            atom_list: ArrayLike,
            chain: Union[int, str] = None
    ):
        """Sets dihedral angles of a single residue in internal coordinates for the atoms defined in atom
        list and in trajectory indices in idxs.

        Parameters
        ----------
        idxs: ArrayLike
            an array in indices corresponding to the progenitor structure in the MolSysIC trajectory.
        dihedrals : ArrayLike
            an array of angles for each idx in ``idxs`` and for each dihedral in ``atom_list``. Should have the shape
            ``(len(idxs), len(atom_list))``
        resi : int
            Residue number of the site being altered
        atom_list : ArrayLike
            Names or array of names of atoms involved in the dihedrals
        chain : int, str, optional
            Chain containing the atoms that are defined by the dihedrals that are being set. Only necessary when there
            is more than one chain in the proteinIC object.

        Returns
        -------
        z_matrix : ArrayLike
            the z_matrix for every frame in ``idxs`` with the new dihedral angel values defined in ``dihedrals``
        """
        chain = self._check_chain(chain)

        dihedrals = np.atleast_2d(dihedrals).T
        atom_list = np.atleast_2d(atom_list)
        z_matrix = self.trajectory.coordinate_array[idxs]

        for i, (values, atoms) in enumerate(zip(dihedrals, atom_list)):
            if (tag := (chain, resi, atoms[1], atoms[2])) not in self.chain_res_name_map:
                raise RuntimeError(f'{atoms} is not a recognized dihedrals of residue {resi}. Please make sure you '
                                   f'have the correct residue number and atom names.')

            pert_idxs = self.chain_res_name_map[tag] + self.chain_res_name_map.get((chain, resi, atoms[2], atoms[1]), [])
            for idx in pert_idxs:
                protein_atom_names = self.z_matrix_names[idx][::-1]
                if tuple(protein_atom_names) == tuple(atoms):
                    deltas = z_matrix[:, idx, 2] - values
                    break
            else:
                raise RuntimeError('')

            z_matrix[:, pert_idxs, 2] -= deltas[:, None]

        return z_matrix

    def get_dihedral(self, resi: int, atom_list: ArrayLike, chain: Union[int, str] = None):
        """Get the dihedral angle(s) of one or more atom sets at the specified residue. Dihedral angles are returned in
        radians

        Parameters
        ----------
        resi : int
            Residue number of the site being altered
        atom_list : ArrayLike
            Names or array of names of atoms involved in the dihedral(s)
        chain : int, str
             Chain identifier. required if there is more than one chain in the protein. Default value = None

        Returns
        -------
        angles: numpy.ndarray
            Array of dihedral angles corresponding to the atom sets in atom list.
        """
        if len(atom_list) == 0:
            return np.array([])

        chain = self._check_chain(chain)

        atom_list = np.atleast_2d(atom_list)
        dihedral_idxs = []
        for atoms in atom_list:
            if (tag := (chain, resi, *atoms)) not in self.topology.dihedrals_by_resnum:
                raise RuntimeError(f'{atoms} is not a recognized dihedral of chain {chain} and residue {resi}. Please '
                                   f'make sure you have the correct residue number and atom names. Note that chilife '
                                   f'considers dihedrals as directional so you may want to try the reverse dihedral.')

            dihedral_idxs.append(list(self.topology.dihedrals_by_resnum[tag]))

        dihedral_idxs = np.array(dihedral_idxs).T
        dihedral_values = self.coords[dihedral_idxs]
        dihedrals = get_dihedrals(*dihedral_values)
        return dihedrals[0] if len(dihedrals) == 1 else dihedrals

    def get_z_matrix_idxs(self, resi: int, atom_list: ArrayLike, chain: Union[int, str] = None):
        """Get the z-matrix indices of the dihedral angle(s) defined by ``atom_list`` and the specified residue and
        chain. Dihedral angles are returned in radians.

        Parameters
        ----------
        resi : int
            Residue number of the site being altered
        atom_list : ArrayLike
            Names or array of names of atoms involved in the dihedral(s)
        chain : int, str
             Chain identifier. required if there is more than one chain in the protein. Default value = None

        Returns
        -------
        dihedral_idxs: numpy.ndarray
            Array of dihedral angles corresponding to the atom sets in atom list.
        """

        if len(atom_list) == 0:
            return np.array([])

        chain = self._check_chain(chain)

        atom_list = np.atleast_2d(atom_list)
        idxs = []
        for atoms in atom_list:
            if (tag := (chain, resi, atoms[1], atoms[2])) not in self.chain_res_name_map:
                raise RuntimeError(f'{atoms} is not a recognized dihedral of chain {chain} and residue {resi}. Please '
                                   f'make sure you have the correct residue number and atom names. Note that chilife '
                                   f'considers dihedrals as directional so you may want to try the reverse dihedral.')

            pert_idxs = self.chain_res_name_map[tag]
            for idx in pert_idxs:
                protein_atom_names = self.z_matrix_names[idx][::-1]
                if tuple(protein_atom_names) == tuple(atoms):
                    idxs.append(idx)
                    break
            else:
                raise RuntimeError('')

        return idxs[0] if len(idxs) == 1 else np.array(idxs)

    def has_clashes(self, distance: float = 1.5) -> bool:
        """Checks for an internal clash between non-bonded atoms.

        Parameters
        ----------
        distance : float
            Minimum distance allowed between non-bonded atoms in angstroms. (Default value = 1.5).

        Returns
        -------
        has_clashes : bool
            True if there are any atoms within ``distance`` angstroms from each other that would constitute a clash
            otherwise False
        """
        diff = self.coords[self.nonbonded[:, 0]] - self.coords[self.nonbonded[:, 1]]
        dist = np.linalg.norm(diff, axis=1)
        has_clashes = np.any(dist < distance)
        return has_clashes

    def _check_chain(self, chain):
        if chain is None and len(self.chains) == 1:
            chain = self.chains[0]

        elif chain is None and len(self.chains) > 1:
            raise ValueError("You must specify the protein chain")

        return chain

    def phi_idxs(self, resnums: ArrayLike = None, chain: Union[int, str] = None):
        """
        Method to return the Z-matrix indices of Phi backbone dihedrals

        Parameters
        ----------
        resnums: int, ArrayLike[int]
            Residues for which to return the Phi value index of the z-matrix
        chain: int, str
            Chain corresponding to the resnums of interest

        Returns
        -------
        idxs: ndarray[int]
            Array of indices corresponding to the phi dihedral angles of the selected residues on the chain
        """
        chain = self._check_chain(chain)
        mask = (self.atom_names == 'C') * (self.atom_chains == chain)
        if resnums is not None:
            resnums = np.atleast_1d(resnums)
            mask *= np.isin(self.atom_resnums, resnums)

        idxs = np.argwhere(mask).flatten()
        idxs = np.array([idx for idx in idxs if idx not in self.nan_idxs])
        return idxs

    def psi_idxs(self, resnums: ArrayLike = None, chain: Union[int, str] = None):
        """
        Method to return the Z-matrix indices of Psi backbone dihedrals

        Parameters
        ----------
        resnums: int, ArrayLike[int]
            Residues for which to return the Psi value index of the z-matrix
        chain: int, str
            Chain corresponding to the resnums of interest

        Returns
        -------
        idxs: ndarray[int]
            Array of indices corresponding to the Psi dihedral angles of the selected residues on the chain
        """
        chain = self._check_chain(chain)
        mask = (self.atom_names == 'N') * (self.atom_chains == chain)
        if resnums is not None:
            resnums = np.atleast_1d(resnums) + 1
            mask *= np.isin(self.atom_resnums, resnums)

        idxs = np.argwhere(mask).flatten()
        idxs = np.array([idx for idx in idxs if idx not in self.nan_idxs])

        return idxs

    def omega_idxs(self, resnums: ArrayLike = None, chain: Union[int, str] = None):
        """
        Method to return the Z-matrix indices of Psi backbone dihedrals

        Parameters
        ----------
        resnums: int, ArrayLike[int]
            Residues for which to return the Psi value index of the z-matrix
        chain: int, str
            Chain corresponding to the resnums of interest

        Returns
        -------
        idxs: ndarray[int]
            Array of indices corresponding to the Psi dihedral angles of the selected residues on the chain
        """

        chain = self._check_chain(chain)
        mask = (self.atom_names == 'CA') * (self.atom_chains == chain)
        if resnums is not None:
            resnums = np.atleast_1d(resnums) + 1
            mask *= np.isin(self.atom_resnums, resnums)

        idxs = np.argwhere(mask).flatten()
        idxs = np.array([idx for idx in idxs if idx not in self.nan_idxs])

        return idxs

    def chi_idxs(self, resnums: ArrayLike = None, chain: Union[int, str] = None):
        """
         Create a list of index arrays corresponding the Z-matrix indices of flexible side chain (chi) dihedrals. Note
         that only chi dihedrals defined in ``chilife.dihedral_defs`` will be  considered.

        Parameters
        ----------
        resnums: int, ArrayLike[int]
            Residues for which to return the Psi value index of the z-matrix
        chain: int, str
            Chain corresponding to the resnums of interest

        Returns
        -------
        chi_idxs: List[np.ndarray]
            list of arrays containing the indices of the z-matrix dihedrals corresponding to the chi dihedrals. Each
            array in the list refers to a specific dihedral: ``[χ1, χ2, ... χn]``. because different residues will
            have different numbers of chi dihedrals the arrays will not be the same size, but they can be concatenated
            using numpy: ``chi_idxs = np.concatenate(IC.chi_idxs(...)))``.

        """
        chain = self._check_chain(chain)
        mask = self.atom_chains == chain
        mask *= ~np.isin(self.atom_names,['N', 'CA', 'C', 'O']) * (self.atom_types != 'H')

        if resnums is not None:
            resnums = np.atleast_1d(resnums)
            mask *= np.isin(self.atom_resnums, resnums)
        else:
            resnums = self.resnums

        chi_idxs = []
        for resnum in resnums:
            tmask = (self.atom_resnums == resnum)
            resname = self.atom_names[tmask][0]
            if resname in dihedral_defs:
                atom_names = [x[-1] for x in dihedral_defs[resname]]
                res_chi_idxs = np.argwhere(tmask * np.isin(self.atom_names, atom_names)).flatten()
            else:
                tmask *= mask
                res_chi_idxs = []
                taken = []
                for idx in np.argwhere(tmask).flat:
                    ddef_names = self.z_matrix_names[idx]
                    ddef_idxs = self.z_matrix_idxs[idx]
                    not_cycle = len(self.topology.graph.get_all_simple_paths(ddef_idxs[1], ddef_idxs[2], 8)) < 2
                    if (ddef_names[1] != 'CA') and (ddef_idxs[1] not in taken) and not_cycle:
                        res_chi_idxs.append(idx)
                        taken.append(ddef_idxs[1])

            chi_idxs.append(res_chi_idxs)

        return chi_idxs

    def __len__(self):
        return len(self.trajectory)

    def load_new(self, z_matrix, **kwargs):
        """
        Replace the current z-matrix ensemble/trajectory with new one.

        Parameters
        ----------
        z_matrix : np.ndarray
            New z-matrix to load into the MolSysIC
        kwargs : dict
            Additional keyword arguments.

            op : dict
                New chain operators for the new z-matrix.
        """
        self.trajectory.load_new(coordinates=z_matrix)
        cart_coords = batch_ic2cart(self.z_matrix_idxs[:, 1:], z_matrix)
        self.protein.trajectory.load_new(cart_coords)

        if 'op' in kwargs:
            op = kwargs['op']
            self.has_chain_operators = True

            if isinstance(op, list):
                # check if all chain operators are virtually the same
                OP0 = op[0]
                for iop in op:
                    scores = [np.abs(OP0[chain][key] - iop[chain][key]).max() for chain in iop for key in iop[chain]]
                    if any( x > 1e-3 for x in scores):
                        break
                else:
                    # if they are we only need the first copy of op
                    op = [op[0]]

                self.chain_operators = op

        elif isinstance(self._chain_operators, list):
            warnings.warn('You are loading in a new internal coordinate trajectory for an internal coordinates object '
                          'that has unique translations and rotations for each frame. These translations and rotations '
                          'will be lost, which can be particularly detrimental for systems with multiple chains. New '
                          'translations & rotations can be applied using the `co` keyword argument.')
            self.chain_operators = None
            self.has_chain_operators = False
        else:
            self.apply_chain_operators()

    def apply_chain_operators(self, idx=None, from_list=False):
        """
        Apply chain operators to the specified frames (``idx``) of the MolSysIC trajectory. If no ``idx`` is provided
        then all chain operators will be applied to all frames.
        Parameters
        ----------
        from_list : bool
            A boolean argument to specify whether each frame previously had its own chain operators.

        idx : int, Array
            Frames or array of frames that should have the chain operators applied.
        """

        idx = np.arange(len(self._chain_operators), dtype=int) if idx is None else idx
        idx = np.atleast_1d(idx)

        cart_coords = self.protein.trajectory.coordinate_array
        if isinstance(self._chain_operators, list):
            for i, op in zip(idx, self._chain_operators[idx]):
                for start, stop in self._chain_segs:
                    current_mx, current_ori = ic_mx(*cart_coords[i, start:start+3])
                    mx = self.chain_operators[start]['mx']
                    ori = self.chain_operators[start]['ori']
                    m2m3 = current_mx @ mx
                    cart_coords[i, start:stop] = (cart_coords[i, start:stop] - current_ori) @ m2m3 + ori

        elif from_list:
            for i in idx:
                for start, stop in self._chain_segs:
                    current_mx, current_ori = ic_mx(*cart_coords[i, start:start+3])
                    mx = self.chain_operators[start]['mx']
                    ori = self.chain_operators[start]['ori']
                    m2m3 = current_mx @ mx
                    cart_coords[i, start:stop] = (cart_coords[i, start:stop] - current_ori) @ m2m3 + ori

        elif isinstance(self._chain_operators, dict):
            for start, end in self._chain_segs:
                current_mx, current_ori = ic_mx(*cart_coords[0, start:start + 3])
                mx = self.chain_operators[start]['mx']
                ori = self.chain_operators[start]['ori']
                m2m3 = current_mx.T @ mx
                cart_coords[:, start:end] = np.einsum('ijk,kl->ijl', cart_coords[:, start:end] - current_ori, m2m3) + ori


    def use_frames(self, idxs):
        """
        Remove all frames except those specified by `idxs`
        Parameters
        ----------
        idxs : int, ArrayLike
            Index or array if indices to keep
        """
        self.trajectory.load_new(coordinates=self.trajectory.coordinate_array[idxs])
        self.protein.load_new(self.protein.trajectory.coordinate_array[idxs])

    def __iter__(self):
        for ts in self.trajectory:
            yield self

    def set_cartesian_coords(self, coords, mask):
        """
        Alter the coordinates of the current frame by assigning new cartwsian coordinates to the atoms set to True and
        ``mask``.

        Parameters
        ----------
        coords : np.ndarray
            Cartesian coordinates to apply to the atoms defined by ``mask`` of the current frame.
        mask : np.ndarray
            Array of bool values defining which atoms of the MolSysIC are being set.
        """
        nrots = len(coords)
        cpy = np.tile(self.protein.trajectory.coordinate_array[0].copy(), (nrots, 1, 1))

        cpy[:, mask] = coords

        z_mats = []
        ops = []
        for submat in cpy:
            z, op = get_z_matrix(submat, self.z_matrix_idxs, self.chain_operator_idxs)
            z_mats.append(z)
            ops.append(op)

        z_mats = np.array(z_mats)
        zcpy = np.tile(self.trajectory.coordinate_array[0].copy(), (nrots, 1, 1))
        zcpy[:, mask] = z_mats[:, mask]

        self.load_new(zcpy, op=ops)

    def shift_resnum(self, delta: int):
        """
        Alter the residue number of all residues in the MolSysIC by adding ``delta``.

        Parameters
        ----------
        delta : int
            Amount to shift residue numbers by. Can be positive or negative.
        """
        self.resnums += delta
        self.atom_resnums += delta

        self.chain_res_name_map = defaultdict(list)
        idxs, b2s, b1s, _ = self.z_matrix_idxs[self.non_nan_idxs].T
        chains = self.atoms[b2s].segids
        resnums = self.atoms[b2s].resnums
        [self.chain_res_name_map[(chain, res, b1, b2)].append(idx)
         for chain, res, b1, b2, idx in
         zip(chains, resnums, self.atom_names[b1s], self.atom_names[b2s], idxs)]

        self.chain_res_name_map = {k: v for k, v in self.chain_res_name_map.items()}

        self.topology.update_resnums()

    @property
    def z_matrix_names(self):
        if not hasattr(self, '_z_matrix_names'):
            self._z_matrix_names = np.array([self.atom_names[[x for x in y if x >= 0]].tolist()
                                             for y in self.z_matrix_idxs], dtype=object)

        return self._z_matrix_names

def reconfigure_cap(cap, atom_idxs, bonds):
    """
    Helper function to reconfigure the "cap" region of bifunctional labels so that the internal coordinates are defined
    correctly.

    Parameters
    ----------
    cap : ArrayLike
        Indices of the cap atoms in the molecule with atoms defined in atom_idxs.
    atom_idxs : ArrayLike
        Indices of all atoms to be considered for reconfiguration containing the cap.
    bonds : ArrayLike
        Array of tuples defining the bonds between atoms in atom_idxs.

    Returns
    -------
    atom_idxs : ArrayLike
        Array of atom indices rearranged so that the cap is terminal to the rest of the atoms.
    """
    atom_idxs = atom_idxs[~np.isin(atom_idxs, cap)]

    # Get bonds to all cap atoms
    sub_bonds = [tuple(bond) for bond in bonds if np.any(np.isin(bond, cap))]

    # Identify bound atoms outside the cap and use the first atom as root
    root = min([bnd[0] for bnd in sub_bonds if bnd[0] not in cap] +
               [bnd[1] for bnd in sub_bonds if bnd[1] not in cap])

    G = ig.Graph(edges=bonds)
    nodes, _, parents = G.bfs(root)

    edges = [(parents[node], node) for node in nodes if node != root]

    for edge in edges:
        mask = np.all(bonds == edge[::-1], axis=1)
        if np.any(mask):
            bndidx = np.argwhere(mask).flat[0]
            bonds[bndidx] = edge

    cap_idxs = [edge[1] for edge in edges if edge[1] not in atom_idxs]
    atom_idxs = np.concatenate((atom_idxs, cap_idxs))

    return atom_idxs


def zmatrix_idxs_to_local(zmatrix_idxs):
    """
    Convert Z-matrix indices so that they reference the indices of the z-matrix rather than the parental atom selection
    that was used to construct the z-matrix and may have an index offset or missing atoms.

    Parameters
    ----------
    zmatrix_idxs : ArrayLike
        z-matrix indices

    Returns
    -------
    zmatrix_idxs : ArrayLike
        Array of indices modified for internal reference.
    """
    idxmap = {d[-1]: i for i, d in enumerate(zmatrix_idxs)}
    new_zmatrix_idxs = []
    for d in zmatrix_idxs:
        d = [idxmap[di] for di in d]
        if (dl := len(d)) < 4:
            d = [np.nan for i in range(4 - dl)] + d
        new_zmatrix_idxs.append(d[::-1])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        new_z_matrix_idxs = np.array(new_zmatrix_idxs).astype(int)

    return new_z_matrix_idxs


def get_chainbreak_idxs(z_matrix_idxs):
    """
    Get indices of atoms that  chain breaks
    Parameters
    ----------
    z_matrix_idxs: ArrayLike
        index map of z_matrix
    Returns
    -------
    chainbreak_idxs:
        indices of atoms that start new chains
    """
    chainbreak_idxs = []
    for idxs in z_matrix_idxs:
        if np.sum(idxs < 0) == 3:
            chainbreak_idxs.append(idxs[0])

    return chainbreak_idxs


def ic_mx(*p: ArrayLike) -> Tuple[ArrayLike, ArrayLike]:
    """
    Calculates a rotation matrix and translation to transform a set of atoms to global coordinate frame from a local
    coordinated frame defined by ``p``. This function is a wrapper to quickly and easily transform cartesian
    coordinates, made from internal coordinates, to their appropriate location in the global frame.

    Parameters
    ----------
    p : ArrayLike
        Coordinates of the three atoms defining the local coordinate frame of the IC object

    Returns
    -------
    rotation_matrix : np.ndarray
        Rotation  matrix to rotate from the internal coordinate frame to the global frame
    origin : np.ndarray
        New origin position in 3-dimensional space
    """

    p1, p2, p3 = p

    # Define new X axis
    v12 = p2 - p1
    v12 /= np.linalg.norm(v12)

    # Define new Y axis
    v23 = p3 - p2
    p23_x_comp = v23.dot(v12)
    v23 -= p23_x_comp * v12
    v23 /= np.linalg.norm(v23)

    # Define new z axis
    z_axis = np.cross(v12, v23)

    # Create rotation matrix
    rotation_matrix = np.array([v12, v23, z_axis])
    origin = p1

    return rotation_matrix, origin


def get_z_matrix(coords, z_matrix_idxs, chain_operator_idxs=None):
    """
    Given a set of cartesian coordinates and indices defining a z-matrix, calcualte the z-matrix.

    Parameters
    ----------
    coords : np.ndarray
        Array of cartesian coordinates.
    z_matrix_idxs : np.ndarray
        Array of indices defining bond-angle-torsions of the z-matrix.
    chain_operator_idxs : np.ndarray
        Array of indices defining different chains.

    Returns
    -------
    z_matrix : np.ndarray
        Z-matrix (internal) coordinates.
    """
    z_matrix = np.zeros((len(z_matrix_idxs), 3))
    bond_mask = ~(z_matrix_idxs[:, 1] < 0)
    bond_values = coords[z_matrix_idxs[bond_mask, 1]] - coords[z_matrix_idxs[bond_mask, 0]]
    bond_values = np.linalg.norm(bond_values, axis=1)

    angle_mask = ~(z_matrix_idxs[:, 2] < 0)
    angle_values = [coords[z_matrix_idxs[angle_mask, i]] for i in range(3)]
    angle_values = get_angles(*angle_values)

    dihedral_mask = ~(z_matrix_idxs[:, 3] < 0)
    dihedral_values = [coords[z_matrix_idxs[dihedral_mask, i]] for i in range(4)]
    dihedral_values = get_dihedrals(*dihedral_values)

    z_matrix[bond_mask, 0] = bond_values
    z_matrix[angle_mask, 1] = angle_values
    z_matrix[dihedral_mask, 2] = dihedral_values
    if chain_operator_idxs is None:
        return z_matrix

    chain_operator = {}
    for cidx in chain_operator_idxs:
        chain_operator_def = z_matrix_idxs[cidx + 2]
        if chain_operator_def[-1] < 0 <= chain_operator_def[-2]:
            pos = coords[chain_operator_def[:3]][::-1]
            mx, ori = ic_mx(*pos)
        else:
            mx, ori = np.eye(3), coords[cidx].copy()
        chain_operator[cidx] = {'mx': mx, 'ori': ori}

    return z_matrix, chain_operator



import pickle
import pytest
import MDAnalysis
import chilife as xl
from chilife.Topology import *
from MDAnalysis.topology.guessers import guess_angles, guess_dihedrals

ubq = xl.MolSys.from_pdb('test_data/1ubq.pdb', sort_atoms=True)
mbp = MDAnalysis.Universe("test_data/2klf.pdb", in_memory=True)
ubqu = MDAnalysis.Universe('test_data/1ubq.pdb')


def test_guess_bonds():
    bonds = xl.guess_bonds(mbp.atoms.positions, mbp.atoms.types)
    dist = np.linalg.norm(mbp.atoms.positions[bonds[:, 0]] - mbp.atoms.positions[bonds[:, 1]], axis=1)

    ans = np.load('test_data/bonds.npy')
    ans_dist = np.load('test_data/bond_dist.npy')
    sorted_args = np.lexsort((ans[:, 0], ans[:, 1]))

    np.testing.assert_equal(bonds, ans[sorted_args])
    np.testing.assert_allclose(dist, ans_dist[sorted_args])


@pytest.mark.parametrize('prot', [ubqu, mbp])
def test_get_angle_defs(prot):
    bonds = xl.guess_bonds(prot.atoms.positions, prot.atoms.types)
    prot.add_bonds(bonds)
    graph1 = ig.Graph(n=len(prot.atoms), edges=bonds)
    angles = set(get_angle_defs(graph1))

    angles2 = set(guess_angles(prot.bonds))

    assert angles == angles2


@pytest.mark.parametrize('prot', [ubqu, mbp])
def test_get_dihedral_defs(prot):
    bonds = xl.guess_bonds(prot.atoms.positions, prot.atoms.types)
    prot.add_bonds(bonds)

    graph1 = ig.Graph(n=len(prot.atoms), edges=bonds)
    dihedrals = set(get_dihedral_defs(graph1))

    prot.add_angles(guess_angles(prot.bonds))
    dihedrals2 = set(guess_dihedrals(prot.angles))

    assert dihedrals == dihedrals2


def test_construction():
    bonds = xl.guess_bonds(ubq.coords, ubq.atypes)
    top = Topology(ubq, bonds)
    for key, val in top.dihedrals_by_resnum.items():

        names = tuple(ubq.atoms[idx].name for idx in val)
        assert names == key[2:]

        resnum = ubq.atoms[val[-2]].resnum
        assert resnum == key[1]

        chain = ubq.atoms[val[-1]].segid
        assert chain == key[0]


def test_get_z_matrix():
    mbp_bonds = xl.guess_bonds(mbp.atoms.positions, mbp.atoms.types)
    top = Topology(mbp, mbp_bonds)

    test_idxs = top.get_zmatrix_dihedrals()

    with open('test_data/top_idxs.pkl', 'rb') as f:
        ans_idxs = pickle.load(f)

    for a, b in zip(test_idxs, ans_idxs):
        np.testing.assert_equal(a, b)


def test_dihedrals_by_atom():
    mbp_bonds = xl.guess_bonds(mbp.atoms.positions, mbp.atoms.types)
    top = Topology(mbp, mbp_bonds)


    with open('test_data/mbp_dihedrals.pkl', 'rb') as f:
        ans = pickle.load(f)

    assert top.dihedrals == ans

    with open('test_data/dihedrals_by_atom.pkl', 'rb') as f:
        ans = pickle.load(f)

    assert top.dihedrals_by_atoms == ans


def test_has_rings():
    Y59 = ubq.select_atoms('resnum 59')
    N60 = ubq.select_atoms('resnum 60')
    Y59_top = Topology(Y59, xl.guess_bonds(Y59.positions, Y59.types))
    N60_top = Topology(N60, xl.guess_bonds(N60.positions, N60.types))

    assert Y59_top.has_rings
    assert not N60_top.has_rings


def test_cycle_idx():
    Y59 = ubq.select_atoms('resnum 59')
    N60 = ubq.select_atoms('resnum 60')
    Y59_top = Topology(Y59, xl.guess_bonds(Y59.positions, Y59.types))
    N60_top = Topology(N60, xl.guess_bonds(N60.positions, N60.types))

    np.testing.assert_equal(Y59_top.ring_idxs, [[5, 6, 7, 8, 9, 10]])
    assert N60_top.ring_idxs == []



#! /usr/bin/env python
from pathlib import Path
from zipfile import ZipFile
from io import BytesIO
import numpy as np
from numpy.lib.format import read_array
import chilife as xl
import oldProteinIC as opic
import sys, os
from importlib import reload
import shutil
import argparse
import zipfile

parser = argparse.ArgumentParser(prog='update_rotlibs',
                                 description='Takes old (incompatible) versions of chiLife rotlibs and creates updated '
                                             '(compatible) versions.',
                                 )
parser.add_argument('files_names', nargs='+', help='Names of the old rotlib files you wish to update')
parser.add_argument('-d', '--description', default=None, help='Description text you would like to add to the rotamer '
                                                              'library. Can be a string or a file')
parser.add_argument('-c', '--comment',  default=None, help='Description text you would like to add to the rotamer '
                                                           'library. can be a string or a file')
parser.add_argument('-r', '--reference',  default=None, help='Reference text you would like to add to the rotamer '
                                                             'library can be a string or a file.')

args = parser.parse_args()


def main():

    adding_metadata = ((args.description is not None) or
                       (args.comment is not None) or
                       (args.reference is not None))

    if adding_metadata and len(args.files_names) > 1:
        raise RuntimeError("You can only add descriptions, comments, or references when updating one library at a time")

    description = check_textfile(args.description)
    comment = check_textfile(args.comment)
    reference = check_textfile(args.reference)

    for filename in args.files_names:

        if filename.endswith('_rotlib.npz'):
            shutil.copy(filename, filename + '.bku')
            data = get_data(filename, description, comment, reference)
            np.savez(filename, **data)

        elif filename.endswith('_drotlib.zip'):
            shutil.copy(filename, filename + '.bku')
            libname = filename[:-12]

            with zipfile.ZipFile(filename, 'r') as archive:
                for f in archive.namelist():
                    if 'csts' in f:
                        with archive.open(f) as of:
                            csts = np.load(of)
                            np.save(f'{libname}_csts.npy', csts)
                    elif f[-12] == 'A':
                        with archive.open(f) as of:
                            libA = get_data(of, description, comment, reference)
                            np.savez(f'{libname}A_rotlib.npz', **libA)
                    elif f[-12] == 'B':
                        with archive.open(f) as of:
                            libB = get_data(of, description, comment, reference)
                            np.savez(f'{libname}B_rotlib.npz', **libB)

            with zipfile.ZipFile(f'{libname}_drotlib.zip', mode='w') as archive:
                archive.write(f'{libname}A_rotlib.npz')
                archive.write(f'{libname}B_rotlib.npz')
                archive.write(f'{libname}_csts.npy')

            os.remove(f'{libname}A_rotlib.npz')
            os.remove(f'{libname}B_rotlib.npz')
            os.remove(f'{libname}_csts.npy')

        else:
            raise RuntimeError(f'{filename} is not a valid chiLife rotlib. All valid chiLife rotlibs end with '
                               f'`_rotlib.npz` or `_drotlib.zip`.')



def get_data(filename, description = None, comment = None, reference = None):

    with np.load(filename, allow_pickle=True) as f:
        version = f['format_version']

    if version < 1.2:
        data = get_data_lt_1_2(filename)

    elif version == 1.2:
        data = get_data_1_2(filename)

    else:
        data = {}
        with np.load(filename, allow_pickle=True) as f:
            for key in f:
                data[key] = f[key]

    for key, val in zip(('description', 'comment', 'reference'), (description, comment, reference)):
        if (key not in data) or (val is not None):
            data[key] = val

    if 'backbone_atoms' not in data:
        data['backbone_atoms'] = ["H", "N", "CA", "HA", "C", "O"]
        data['aln_atoms'] = ['N', 'CA', 'C']

    if not hasattr(data['internal_coords'].item().protein, 'icodes'):
        resnames = data['internal_coords'].item().protein.resnames
        icodes = np.array(["" for _ in resnames])
        record_types = np.array(['ATOM' for _ in resnames])


        data['internal_coords'].item().protein.__dict__['icodes'] = icodes
        data['internal_coords'].item().protein.__dict__['record_types'] = record_types

    data['format_version'] = 1.5


    return data

def get_data_lt_1_2(filename):
    sys.modules['chilife.ProteinIC'] = opic
    data = {}
    with np.load(filename, allow_pickle=True) as f:
        for key in f:
            data[key] = f[key]

    IC0 = data['internal_coords'][0]
    bonds = IC0.bonded_pairs
    z_matrix_idxs = IC0.zmat_idxs[1]

    z_matrix = []
    for ic in data['internal_coords']:
        z_matrix.append(ic.zmats[1])

    z_matrix = np.array(z_matrix)
    z_matrix_idxs = np.c_[np.arange(len(z_matrix_idxs)), z_matrix_idxs]
    ic_coords = np.array([ic.coords for ic in data['internal_coords']])

    anames = IC0.atom_names
    atypes = IC0.atom_types
    resnames = np.array([val for val in IC0.resnames.values()])
    resindices = np.array([0 for a in IC0.atoms])
    resnums = IC0.resis.copy()
    segindices = np.array([0])
    segids = ['A']
    uni = xl.make_mda_uni(anames, atypes, resnames, resindices, resnums, segindices, segids)
    uni.load_new(ic_coords)

    del sys.modules['chilife.ProteinIC']
    reload(xl)
    P = xl.MolSysIC(z_matrix, z_matrix_idxs, uni, bonds=bonds)
    data['internal_coords'] = P

    return data

def get_data_1_2(filename):

    with ZipFile(filename, 'r') as fb:
        with fb.open('internal_coords.npy', 'r') as finner:
            ic_bytes = finner.read()

    ic_bytes = ic_bytes.replace(b'chilife.Protein\n', b'chilife.MolSys\n')
    ic_bytes = ic_bytes.replace(b'\nProtein\n', b'\nMolSys\n')
    ic_bytes = ic_bytes.replace(b'chilife.ProteinIC\nProteinIC', b'chilife.MolSysIC\nMolSysIC')

    data = {}
    with np.load(filename, allow_pickle=True) as f:
        for key in f:
            if key != 'internal_coords':
                data[key] = f[key]

    data['internal_coords'] = read_array(BytesIO(ic_bytes), allow_pickle=True)

    return data

def check_textfile(file):

    if Path(str(file)).exists():

        with open(file, 'r') as f:
            lines = f.readlines()

        lines = "\n".join(lines)
    else:
        lines = file

    return lines

main()


import numpy as np
import matplotlib.pyplot as plt
import chilife

#
# def test_generation():
#     mv0 = ProEPR.fetch('2mv0')
#     SLTraj = ProEPR.SpinLabelTraj('R1A', 45, 'A', mv0)
#
#
# def test_iter():
#     mv0 = ProEPR.fetch('2mv0')
#     SLTraj = ProEPR.SpinLabelTraj('R1A', 45, 'A', mv0)
#     for SL in SLTraj:
#         print(SL)
#
#
# def test_zip():
#     mv0 = ProEPR.fetch('2mv0')
#     SLTraj1 = ProEPR.SpinLabelTraj('R1A', 45, 'A', mv0)
#     SLTraj2 = ProEPR.SpinLabelTraj('R1A', 55, 'A', mv0)
#     for SL1, SL2 in zip(SLTraj1, SLTraj2):
#         print(SL1, SL2)
#
#
# def test_get_dd():
#     mv0 = ProEPR.fetch('2mv0')
#     SLTraj1 = ProEPR.SpinLabelTraj('R1A', 45, 'A', mv0)
#     SLTraj2 = ProEPR.SpinLabelTraj('R1A', 55, 'A', mv0)
#     r = np.linspace(10, 80, 256)
#     P = ProEPR.distance_distribution(SLTraj1, SLTraj2, r)
#
#     plt.plot(r, P)
#     plt.show()



import os, hashlib
import pickle
import random
from functools import partial
import numpy as np
import pytest
import MDAnalysis as mda

import chilife

pdbids = ["1ubq", "1a2w", '1az5']
ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
resis = [
    (key, -90, 160)
    for key in chilife.dihedral_defs
    if key
    not in (("R1B", "R1C", "CYR1", "MTN", "ALA", "GLY", 'HID', 'HIE', 'ASH', 'GLH', 'CYM', 'LYN', 'HIP') +
            tuple(chilife.USER_LIBRARIES) +
            tuple(chilife.USER_dLIBRARIES))
]


@pytest.mark.parametrize("res", resis)
def test_read_dunbrack(res):
    res, phi, psi = res

    dlib = chilife.read_bbdep(res, -70, 90)


    with np.load(f"test_data/{res}_{phi}_{psi}.npz", allow_pickle=True) as f:
        dlib_ans = {key: f[key] for key in f if key != "allow_pickle"}

    for key in dlib_ans:
        if dlib_ans[key].dtype not in [np.dtype(f"<U{i}") for i in range(1, 5)]:
            np.testing.assert_almost_equal(dlib_ans[key], dlib[key], decimal=5)
        else:
            assert np.all(np.char.equal(dlib_ans[key], dlib[key]))


def test_sort_pdb():
    pdbfile = "test_data/trt.pdb"
    x = chilife.sort_pdb(pdbfile)

    linest = [line for line in x if line[76:79].strip() != 'H']
    test = hashlib.md5("".join(linest).encode("utf-8")).hexdigest()

    with open("test_data/trt_sorted.pdb", "r") as f:
        lines = f.readlines()
        linesa = [line for line in lines if line[76:79].strip() != 'H']
        ans = hashlib.md5("".join(linesa).encode("utf-8")).hexdigest()

    assert test == ans


def test_sort_pdb2():
    x = chilife.sort_pdb("test_data/SL_GGAGG.pdb")

    with open("test_data/SL_GGAGG_tmp.pdb", "w") as f:
        for line in x:
            f.write(line)

    U = mda.Universe("test_data/SL_GGAGG_tmp.pdb", in_memory=True)
    os.remove("test_data/SL_GGAGG_tmp.pdb")

    # check if atom order compatible with internal coordinates system.
    ICs = chilife.MolSysIC.from_atoms(U, preferred_dihedrals=[["C", "N", "CA", "C"]])


def test_sort_H():
    x = chilife.sort_pdb('../src/chilife/data/rotamer_libraries/residue_pdbs/lys.pdb')
    H_names = [xx[12:16] for xx in x[9:]]
    np.testing.assert_equal(H_names, [' H  ', ' HA ', '3HB ', '2HB ',
                                      '3HG ', '2HG ', '3HD ', '2HD ',
                                      '2HE ', '3HE ', '3HZ ', '2HZ ',
                                      '1HZ '])


def test_sort_residue():
    lines = chilife.sort_pdb('test_data/TUM.pdb', aln_atoms=['N3', 'C4', 'C5'])
    with open('test_data/TUM_srtd.pdb', 'r') as f:
        alines = f.readlines()

    test = hashlib.md5("".join(lines[0]).encode("utf-8")).hexdigest()
    ans = hashlib.md5("".join(alines).encode("utf-8")).hexdigest()
    assert test == ans


def test_sort_many_models():
    x = chilife.sort_pdb("test_data/msort.pdb")
    with open("test_data/msort_tmp.pdb", "w") as f:
        for i, struct in enumerate(x):
            f.write(f"MODEL {i}\n")
            f.writelines(struct)
            f.write("ENDMDL\n")

    with open("test_data/msort_tmp.pdb", "r") as f:
        heavy_test = [line for line in f.readlines() if line[76:79].strip() != 'H']

        test = hashlib.md5("".join(heavy_test).encode("utf-8")).hexdigest()

    with open("test_data/msort_ans.pdb", "r") as f:
        heavy_ans = [line for line in f.readlines() if line[76:79].strip() != 'H']

        ans = hashlib.md5("".join(heavy_ans).encode("utf-8")).hexdigest()

    os.remove('test_data/msort_tmp.pdb')

    assert test == ans


def test_mutate():
    protein = mda.Universe("test_data/1ubq.pdb", in_memory=True).select_atoms("protein")
    SL = chilife.SpinLabel("R1C", site=28, protein=protein)

    labeled_protein = chilife.mutate(protein, SL)
    ub1_A28R1 = mda.Universe("test_data/1ubq_A28R1.pdb", in_memory=True)

    np.testing.assert_almost_equal(ub1_A28R1.atoms.positions, labeled_protein.atoms.positions, decimal=3)


def test_mutate2():
    protein = mda.Universe('test_data/1ubq.pdb', in_memory=True).select_atoms("protein")
    SL1 = chilife.SpinLabel("R1C", site=28, protein=protein)
    SL2 = chilife.SpinLabel("R1C", site=48, protein=protein,)

    labeled_protein = chilife.mutate(protein, SL1, SL2)
    ub1_A28R1_K48R1 = mda.Universe("test_data/ub1_A28R1_K48R1.pdb")
    np.testing.assert_almost_equal(ub1_A28R1_K48R1.atoms.positions, labeled_protein.atoms.positions, decimal=3)


def test_mutate3():
    protein = mda.Universe("test_data/1omp_H.pdb").select_atoms("protein")
    SL1 = chilife.SpinLabel("R1C", site=238, protein=protein, use_H=True)
    SL2 = chilife.SpinLabel("R1C", site=311, protein=protein, use_H=True)

    labeled_protein = chilife.mutate(protein, SL1, SL2)
    assert len(labeled_protein.atoms) != len(protein.atoms)


def test_mutate4():
    protein = mda.Universe("test_data/1omp_H.pdb").select_atoms("protein")
    D41G = chilife.RotamerEnsemble("GLY", 41, protein=protein, eval_clash=False)
    S238A = chilife.RotamerEnsemble("ALA", 238, protein=protein, eval_clash=False)
    mPro = chilife.mutate(protein, D41G, S238A, add_missing_atoms=False)
    D41G_pos = mPro.select_atoms("resnum 41").positions
    S238A_pos = mPro.select_atoms("resnum 238").positions
    np.testing.assert_almost_equal(D41G._coords[0], D41G_pos, decimal=6)
    np.testing.assert_almost_equal(S238A._coords[0], S238A_pos, decimal=6)


def test_mutate5():
    PPII = mda.Universe('test_data/PolyProII.pdb')
    TOC = chilife.SpinLabel('TOC', 8, PPII)

    PPIIm = chilife.mutate(PPII, TOC)
    assert PPIIm.residues[-1].resname == 'NHH'

def test_mutate6():
    SL = chilife.SpinLabel("R1C", site=28, protein=ubq)
    labeled_protein = chilife.mutate(ubq, SL, rotamer_index=10)

    # Make sure the newly mutated protein has the coordinates of the rotamer selected
    np.testing.assert_allclose(labeled_protein.select_atoms(SL.selstr).positions, SL.coords[10])

    # Make sure that is not the same as the first rotamer.
    assert np.max(labeled_protein.select_atoms(SL.selstr).positions - SL.coords[0]) > 1

def test_mutate7():
    SL = chilife.SpinLabel("R1C", site=28, protein=ubq)
    labeled_protein = chilife.mutate(ubq, SL, rotamer_index='all')

    assert len(labeled_protein.trajectory) == len(SL)

    label_atoms =labeled_protein.select_atoms(SL.selstr)
    for i, ts in enumerate(labeled_protein.trajectory):
        np.testing.assert_allclose(label_atoms.positions, SL.coords[i])

def test_mutate8():
    SL1 =  chilife.SpinLabel("R1C", site=28, protein=ubq)
    SL2 = chilife.RotamerEnsemble("LYS", site=48, protein=ubq)
    labeled_protein = chilife.mutate(ubq, SL1, SL2, rotamer_index='all')

    assert len(labeled_protein.trajectory) == max(len(SL1), len(SL2))

    for i, ts in enumerate(labeled_protein.trajectory):
        for SL in  (SL1, SL2):
            label_atoms = labeled_protein.select_atoms(SL.selstr)
            if len(SL) <= i:
                np.testing.assert_allclose(label_atoms.positions, SL.coords[-1])
            else:
                np.testing.assert_allclose(label_atoms.positions, SL.coords[i])


def test_mutate9():

    prot = chilife.load_protein("test_data/ubq_ccap.pdb")
    K = chilife.RotamerEnsemble("LYS", site=76, protein=prot, use_H=True)
    mut = chilife.mutate(prot, K)


def test_add_missing_atoms():
    protein = mda.Universe("test_data/1omp.pdb", in_memory=True).select_atoms("protein")
    new_prot = chilife.mutate(protein)
    assert len(new_prot.atoms) != len(protein.atoms)
    assert len(new_prot.atoms) == 2877


@pytest.mark.parametrize(
    "res", set(chilife.dihedral_defs.keys()) -
           {"CYR1", "MTN", "R1M", "R1C", 'HID', 'HIE', 'ASH', 'GLH', 'CYM', 'LYN', 'HIP'} -
           chilife.USER_dLIBRARIES -
           chilife.USER_LIBRARIES
)
def test_sort_and_internal_coords(res):
    pdbfile = chilife.RL_DIR / f"residue_pdbs/{res.lower()}.pdb"
    lines = chilife.sort_pdb(str(pdbfile))
    anames = [line[13:16] for line in lines if "H" not in line[12:16]]

    with open(pdbfile, "r") as f:
        ans = [
            line[13:16]
            for line in f.readlines()
            if line.startswith("ATOM")
            if "H" not in line[12:16]
        ]

    assert anames == ans


def test_get_min_topol():
    with open('test_data/DHC.pdb', 'r') as f:
        lines = f.readlines()

    all_lines = []
    for line in lines:
        if line.startswith('MODEL'):
            newlines = []
        elif line.startswith(('ATOM', 'HETATOM')):
            newlines.append(line)
        elif line.startswith('ENDMDL'):
            all_lines.append(newlines)

    min_bonds = chilife.get_min_topol(all_lines)
    ans = {(0, 1), (0, 2), (0, 6), (2, 3), (2, 4), (2, 5), (6, 7), (6, 16), (7, 8), (7, 10), (7, 17), (8, 9), (8, 23),
           (10, 11), (10, 18), (10, 19), (11, 12), (11, 13), (12, 15), (12, 20), (13, 14), (13, 21), (14, 15), (14, 22),
           (15, 77), (23, 24), (23, 27), (24, 25), (24, 28), (24, 29), (25, 26), (25, 30), (30, 31), (30, 34), (31, 32),
           (31, 35), (31, 36), (32, 33), (32, 37), (37, 38), (37, 41), (38, 39), (38, 42), (38, 43), (39, 40), (39, 44),
           (44, 45), (44, 54), (45, 46), (45, 48), (45, 55), (46, 47), (46, 61), (48, 49), (48, 56), (48, 57), (49, 50),
           (49, 51), (50, 53), (50, 58), (51, 52), (51, 60), (52, 53), (52, 59), (53, 77), (61, 62), (61, 63), (64, 73),
           (64, 77), (65, 75), (66, 75), (66, 77), (67, 73), (68, 71), (68, 72), (68, 74), (68, 77), (69, 76), (70, 76),
           (71, 73), (71, 78), (71, 79), (72, 76), (72, 82), (72, 83), (74, 75), (74, 80), (74, 81)}

    assert min_bonds == ans


def test_sort_nonuni_topol():
    srtd = chilife.sort_pdb('test_data/test_nonuni_top.pdb', uniform_topology=False)

    with open('test_data/nonuni_topol_ans.pkl', 'rb') as f:
        ans = pickle.load(f)

    for test_model, ans_model in zip(srtd, ans):
        for test_line, ans_line in zip(test_model, ans_model):
            assert test_line == ans_line


def test_get_angle():
    sel = ubq.select_atoms('resnum 23 and name N CA C')
    ang = chilife.get_angle(sel.positions)
    assert ang - np.deg2rad(110.30299155404542) < 1e-6


def test_get_angles():
    Ns = ubq.select_atoms('name N').positions
    CAs = ubq.select_atoms('name CA').positions
    Cs = ubq.select_atoms('name C').positions

    angs = chilife.get_angles(Ns, CAs, Cs)
    ans = [chilife.get_angle([N, CA, C]) for N, CA, C in zip(Ns, CAs, Cs)]
    np.testing.assert_almost_equal(angs, ans)


def test_get_dihedral():
    sel = ubq.select_atoms('(resnum 23 and name N CA C) or (resnum 24 and name N)')
    dihe = chilife.get_dihedral(sel.positions)
    assert dihe - np.deg2rad(37.20999032936247) < 1e-6


def test_get_dihedrals():
    N1s = ubq.select_atoms('name N').positions[:-1]
    CAs = ubq.select_atoms('name CA').positions[:-1]
    Cs = ubq.select_atoms('name C').positions[:-1]
    N2s = ubq.select_atoms('name N').positions[1:]

    dihes = chilife.get_dihedrals(N1s, CAs, Cs, N2s)
    ans = [chilife.get_dihedral([N1, CA, C, N2]) for N1, CA, C, N2 in zip(N1s, CAs, Cs, N2s)]
    np.testing.assert_almost_equal(dihes, ans, decimal=6)


def test_guess_mobile_dihedrals():
    label = chilife.MolSys.from_pdb('test_data/TEP.pdb')

    ICs = chilife.MolSysIC.from_atoms(label)
    mbd = chilife.guess_mobile_dihedrals(ICs)

    ans = [['N', 'CA', 'CB', 'CB2'],
           ['CA', 'CB', 'CB2', 'NG'],
           ['ND', 'CE3', 'CZ3', 'C31'],
           ['CZ1', 'C11', 'C12', 'O12'],
           ['C11', 'C12', 'N12', 'C13'],
           ['C12', 'N12', 'C13', 'C14'],
           ['N12', 'C13', 'C14', 'C16'],
           ['C14', 'C15', 'C18', 'C19'],
           ['C14', 'C15', 'C20', 'C21'],
           ['C16', 'C17', 'C22', 'C23'],
           ['C16', 'C17', 'C24', 'C25']]

    np.testing.assert_equal(mbd, ans)


def test_get_site_volume():
    vols = [chilife.get_site_volume(site, ubq) for site in ubq.residues.resnums]
    ans = np.load('test_data/test_get_site_volume.npy')
    np.testing.assert_almost_equal(vols, ans)


def test_make_peptide():
    pep = chilife.make_peptide("ACDEF[R1M]GHIKL<COC1=CC=C(C=C1)CC(C(=O)O)N>MNPQR[R1C]STVWY")

    test = pep.positions.copy()
    ans = np.load('test_data/test_mk_pep.npy')

    ans -= np.mean(ans, axis=0)

    test -= np.mean(test, axis=0)
    mx = np.linalg.inv(test.T @ test) @ test.T @ ans
    test = test @ mx
    diff = test - ans
    rms = np.sqrt(np.sum(diff * diff) / len(test))

    assert rms < 3

def test_make_strand():
    phi = np.ones(22) * -140
    psi = np.ones(22) * 130
    omega = np.ones(22) * -178

    pep = chilife.make_peptide("ACDEF[R1M]GHIKL<COC1=CC=C(C=C1)CC(C(=O)O)N>MNPQR[R1C]STVWY", phi, psi, omega)
    pepIC = chilife.MolSysIC.from_atoms(pep)
    phi_idx = pepIC.phi_idxs(np.arange(2, 23))
    psi_idx = pepIC.psi_idxs(np.arange(1, 22))
    tphi = np.rad2deg(pepIC.z_matrix[phi_idx, 2])
    tpsi  = np.rad2deg(pepIC.z_matrix[psi_idx, 2])
    np.testing.assert_allclose(phi[1:], tphi)
    np.testing.assert_allclose(psi[:-1], tpsi)


def test_make_pep_w_cap():
    pep = chilife.make_peptide("[ACE]A[R1M]A[NME]")
    ans = np.load('test_data/ACE_AR1A_NME.npy')
    np.testing.assert_almost_equal(pep.positions, ans, decimal=4)

def test_make_pep_w_cap2():
    phi = np.array([-61, -63])
    psi = np.array([-40, -42])
    omega = np.array([-176, 176])
    pep = chilife.make_peptide("A[R1M]A", phi=phi, psi=psi, omega=omega)
    ans = np.load('test_data/mkpep2.npy')
    np.testing.assert_almost_equal(pep.positions, ans, decimal=4)


def test_make_pep_w_cap3():
    phi = np.array([-60, -62, -64])
    psi = np.array([-39, -41, -43])
    omega = np.array([-175, 180, 175])

    pep = chilife.make_peptide("[ACE]A[R1M]A[NME]", phi=phi, psi=psi, omega=omega)
    ans = np.load('test_data/test_mkpep3.npy')
    np.testing.assert_almost_equal(pep.positions, ans, decimal=4)





def test_mkpep_template():
    ATII = chilife.fetch('6JOD').select_atoms('segid B')
    pep = chilife.make_peptide("DRV[TOC]IHPF", template=ATII)

    ATII_IC = chilife.MolSysIC.from_atoms(ATII)
    phi_idx = ATII_IC.phi_idxs(np.arange(2, 23))
    psi_idx = ATII_IC.psi_idxs(np.arange(1, 22))
    phi = np.rad2deg(ATII_IC.z_matrix[phi_idx, 2])
    psi  = np.rad2deg(ATII_IC.z_matrix[psi_idx, 2])

    pep_IC = chilife.MolSysIC.from_atoms(pep)
    phi_idx = pep_IC.phi_idxs(np.arange(2, 23))
    psi_idx = pep_IC.psi_idxs(np.arange(1, 22))
    tphi = np.rad2deg(pep_IC.z_matrix[phi_idx, 2])
    tpsi  = np.rad2deg(pep_IC.z_matrix[psi_idx, 2])

    np.testing.assert_allclose(phi, tphi)
    np.testing.assert_allclose(psi, tpsi)


def test_parsed_sequence():
    pseq = chilife.parse_sequence("[ACE]AaNIE<cccn>[NHH]")
    for a, b in zip(pseq, ['ACE', 'ALA', 'ALA', 'ASN', 'ILE', 'GLU', 'cccn', 'NHH']):
        assert a == b


s2rkeys = ["COC1=CC=C(C=C1)CC(C(=O)O)N",
           "CC1CC=NC1C(=O)NCCCCC(C(=O)O)N",
           "CC(C)(C)OC(=O)N[C@@H](CCCCN)C(O)=O",
           "C(CC1=CC=C(C=C1)OCCCCC#C)(C(=O)O)N"]
s2rans = [f"test_data/s2rm{i+1}.npy" for i in range(len(s2rkeys))]
@pytest.mark.parametrize('key, ans', zip(s2rkeys, s2rans))
def test_smiles2residue(key, ans):
    m1 = chilife.smiles2residue(key, randomSeed=0)
    test_mask = np.argsort(m1.names)
    test = m1.positions[test_mask]

    answer = np.load(ans)
    answer -= np.mean(answer, axis=0)

    test -= np.mean(test, axis=0)
    mx = np.linalg.inv(test.T @ test) @ test.T @ answer
    test = test @ mx
    diff = test - answer
    rms = np.sqrt(np.sum(diff * diff) / len(test))

    assert rms < 3


def test_smiles2residue2():
    m1 = chilife.smiles2residue("C1C[C@H]2[C@@H]([C@@H]2C1)C(=O)C[C@H]([NH3+])C(=O)[O-]", randomSeed=0)
    test_mask = np.argsort(m1.names)
    test = m1.positions[test_mask]

    answer = np.load('test_data/s2rm5.npy')
    answer -= np.mean(answer, axis=0)

    test -= np.mean(test, axis=0)
    mx = np.linalg.inv(test.T @ test) @ test.T @ answer
    test = test @ mx
    diff = test - answer
    rms = np.sqrt(np.sum(diff * diff) / len(test))

    assert rms < 3

def test_smiles2residue3():
    m1 = chilife.smiles2residue("CCc1c([nH+]c(nc1NC[C@@H](C(=O)[O-])[NH3+])C)C")
    assert True


def test_append_cap():
    pep = chilife.MolSys.from_pdb('test_data/test_make_peptide.pdb')
    new_pep = chilife.append_cap(pep, 'ace')
    new_pep = chilife.append_cap(new_pep, 'nme')

    ace_coords = np.array([[-0.62455994,  0.61491044,  0.946],
                           [ 0.02253714,  1.18988467,  1.832],
                           [-2.13832676,  0.62232832,  0.958],
                           [-2.5042362 ,  1.63777023,  0.907],
                           [-2.51699797,  0.07177849,  0.109],
                           [-2.50306232,  0.16268392,  1.864]])

    assert new_pep.residues[0].resname == 'ACE'
    assert new_pep.residues[0].resnum == 0
    np.testing.assert_allclose(new_pep.residues[0].positions, ace_coords)

    nhh_coords = np.array([[21.63728595, 20.70101107, 18.63843541],
                           [21.81199224, 22.11387508, 18.92988522],
                           [20.84481362, 20.39311986, 18.09318996],
                           [21.42355845, 22.33956019, 19.91116725],
                           [22.86200865, 22.37037609, 18.90219429],
                           [21.28635322, 22.70933948, 18.19786284]])


    assert new_pep.residues[-1].resname == 'NME'
    assert new_pep.residues[-1].resnum == 24
    np.testing.assert_allclose(new_pep.residues[-1].positions, nhh_coords)


def test_add_cap():
    chilife.store_cap('ZZZ', 'test_data/ACE_A_NME.pdb', 'N')
    chilife.store_cap('XXX', 'test_data/ACE_A_NME.pdb', 'C')

    assert (chilife.RL_DIR / f'cap_pdbs/XXX.pdb').exists()
    assert (chilife.RL_DIR / f'cap_pdbs/ZZZ.pdb').exists()

    os.remove(chilife.RL_DIR / f'cap_pdbs/XXX.pdb')
    os.remove(chilife.RL_DIR / f'cap_pdbs/ZZZ.pdb')


import numpy as np
from .RotamerEnsemble import RotamerEnsemble
import chilife


class FluorLabel(RotamerEnsemble):
    """
    Base object for fluoresent labeling experiments.

    Parameters
    ----------
        :param label: str
            Name of desired label, e.g. .
        :param site: int
            MolSys residue number to attach spin label to.
        :param chain: str
            MolSys chain identifier to attach spin label to.
        :param protein: MDAnalysis.Universe, MDAnalysis.AtomGroup
            Object containing all protein information (coords, atom types, etc.)
        :param protein_tree: cKDtree
            k-dimensional tree object associated with the protein coordinates.
    """

    def __init__(self, label, site=None, protein=None, chain=None, **kwargs):
        """
        Create new SpinLabel object.

        :param label: str
            Name of desired spin label, e.g. R1A.
        :param site: int
            MolSys residue number to attach spin label to.
        :param chain: str
            MolSys chain identifier to attach spin label to.
        :param protein: MDAnalysis.Universe, MDAnalysis.AtomGroup
            Object containing all protein information (coords, atom types, etc.)
        :param protein_tree: cKDtree
            k-dimensional tree object associated with the protein coordinates.
        """
        # Overide RotamerEnsemble default of not evaluating clashes
        kwargs.setdefault("eval_clash", True)
        super().__init__(label, site, protein=protein, chain=chain, **kwargs)

        self.label = label

        # Parse important indices
        self.fluor_idx = np.argwhere(
            np.isin(self.atom_names, chilife.FLUOR_ATOMS[label[:3]])
        )

    @property
    def fluor_coords(self):
        """get the fluorescent coordinates of the RotamerEnsemble"""
        return self._coords[:, self.fluor_idx, :].mean(axis=1)[:, 0]

    @property
    def spin_centroid(self):
        return np.average(self.fluor_coords, weights=self.weights, axis=0)

    @classmethod
    def from_mmm(cls, label, site, protein=None, chain=None, **kwargs):
        """
        Create a SpinLabel object using the default MMM protocol with any modifications passed via kwargs
        """
        MMM_maxdist = {
            "R1M": 9.550856367392733,
            "R7M": 9.757254987175209,
            "V1M": 8.237071322458029,
            "M1M": 8.985723827323680,
            "I1M": 12.952083029729994,
        }

        # Store the force field parameter set being used before creating the spin label
        curr_lj = chilife.using_lj_param
        user_lj = kwargs.pop("lj_params", "uff")
        # Set MMM defaults or user defined overrides
        chilife.set_lj_params(user_lj)

        clash_radius = kwargs.pop("clash_radius", MMM_maxdist[label] + 4)
        alignment_method = kwargs.pop("alignment_method", "mmm")
        clash_ori = kwargs.pop("clash_ori", "CA")
        energy_func = kwargs.pop(
            "energy_func", partial(chilife.get_lj_energy, cap=np.inf)
        )
        use_H = kwargs.pop("use_H", True)
        forgive = kwargs.pop("forgive", 0.5)

        # Calculate the SpinLabel
        SL = chilife.SpinLabel(
            label,
            site,
            protein,
            chain,
            alignment_method=alignment_method,
            clash_radius=clash_radius,
            clash_ori=clash_ori,
            energy_func=energy_func,
            use_H=use_H,
            forgive=forgive,
            **kwargs,
        )

        # restore the force field parameter set being used before creating the spin label
        chilife.set_lj_params(curr_lj)
        return SL

    @classmethod
    def from_wizard(
        cls,
        label,
        site=1,
        protein=None,
        chain=None,
        to_find=200,
        to_try=10000,
        vdw=2.5,
        clashes=5,
        **kwargs,
    ):
        prelib = cls(label, site, protein, chain, eval_clash=False, **kwargs)
        if not kwargs.setdefault("use_prior", False):
            prelib.sigmas = np.array([])

        coords = np.zeros((to_find, len(prelib.atom_names), 3))
        internal_coords = []
        i = 0
        if protein is not None:
            protein_clash_idx = prelib.protein_tree.query_ball_point(
                prelib.centroid(), 19.0
            )
            protein_clash_idx = [
                idx for idx in protein_clash_idx if idx not in prelib.clash_ignore_idx
            ]

        a, b = [list(x) for x in zip(*prelib.non_bonded)]
        for _ in range(np.rint(to_try / to_find).astype(int)):
            sample, _, internal_sample = prelib.sample(
                n=to_find, off_rotamer=True, return_dihedrals=True
            )

            # Evaluate internal clashes
            dist = np.linalg.norm(sample[:, a] - sample[:, b], axis=2)
            sidx = np.atleast_1d(np.squeeze(np.argwhere(np.all(dist > 2, axis=1))))
            if len(sidx) == 0:
                continue

            sample = sample[sidx, ...]
            internal_sample = internal_sample[sidx]
            if protein is not None:
                # Evaluate external clashes
                dist = cdist(
                    sample[:, prelib.side_chain_idx].reshape(-1, 3),
                    prelib.protein_tree.data[protein_clash_idx],
                )
                dist = dist.reshape(len(sidx), -1)
                nclashes = np.sum(dist < vdw, axis=1)
                sidx = np.atleast_1d(np.squeeze(np.argwhere(nclashes < clashes)))
            else:
                sidx = np.arange(len(sample))

            if len(sidx) == 0:
                continue

            if i + len(sidx) >= to_find:
                sidx = sidx[: to_find - i]

            coords[i : i + len(sidx)] = sample[sidx]
            internal_coords.append(internal_sample[sidx])
            i += len(sidx)

            if i == to_find:
                break

        coords = coords[coords.sum(axis=(1, 2)) != 0]
        prelib.internal_coords = (
            np.concatenate(internal_coords) if len(internal_coords) > 0 else []
        )
        prelib._dihedrals = np.array(
            [IC.get_dihedral(1, prelib.dihedral_atoms) for IC in prelib.internal_coords]
        )
        prelib._coords = coords
        prelib.weights = np.ones(len(coords))
        prelib.weights /= prelib.weights.sum()
        return prelib



import MDAnalysis as mda
from pathlib import Path
import chilife
import pickle

pdb_directory = Path('../residue_pdbs')
residue_pdbs = pdb_directory.glob('*.pdb')

for pdb in residue_pdbs:

    new_path = Path(pdb.stem + '_ic.pkl')
    if new_path.exists():
        continue

    print(pdb.stem)

    srtd = chilife.sort_pdb(pdb)
    with open(pdb, 'w') as f:
        f.writelines(srtd)

    struct = mda.Universe(str(pdb), in_memory=True)
    pref_d = chilife.dihedral_defs[pdb.stem.upper()]
    print(pref_d)
    ICs = chilife.MolSysIC.from_atoms(struct, preferred_dihedrals=pref_d)
    with open(pdb.stem + '_ic.pkl', 'wb') as f:
        pickle.dump(ICs, f)